summaryrefslogtreecommitdiff
path: root/core/src/xmake/binutils/macho
diff options
context:
space:
mode:
authorruki <[email protected]>2025-12-19 22:55:22 +0800
committerruki <[email protected]>2025-12-19 22:55:22 +0800
commit2c497bc57fc27af645afbf3de2e4e6068ea853dc (patch)
treedef81c9e79c644381502f0bfe7c0ee7547536241 /core/src/xmake/binutils/macho
parentbbe585b9b8927d71f79f8c723110224e31348381 (diff)
add rpath list in binutils
Diffstat (limited to 'core/src/xmake/binutils/macho')
-rw-r--r--core/src/xmake/binutils/macho/deplibs.c2
-rw-r--r--core/src/xmake/binutils/macho/prefix.h39
-rw-r--r--core/src/xmake/binutils/macho/rpath.c146
3 files changed, 175 insertions, 12 deletions
diff --git a/core/src/xmake/binutils/macho/deplibs.c b/core/src/xmake/binutils/macho/deplibs.c
index 40d19b5cb..286a40d20 100644
--- a/core/src/xmake/binutils/macho/deplibs.c
+++ b/core/src/xmake/binutils/macho/deplibs.c
@@ -104,7 +104,7 @@ tb_bool_t xm_binutils_macho_deplibs(tb_stream_ref_t istream, tb_hize_t base_offs
}
xm_binutils_macho_swap_load_command(&lc, swap);
- /* check for LC_LOAD_DYLIB, LC_LOAD_WEAK_DYLIB, LC_REEXPORT_DYLIB, LC_ID_DYLIB */
+ // check for LC_LOAD_DYLIB, LC_LOAD_WEAK_DYLIB, LC_REEXPORT_DYLIB, LC_ID_DYLIB
if (lc.cmd == XM_MACHO_LC_LOAD_DYLIB || lc.cmd == XM_MACHO_LC_ID_DYLIB ||
lc.cmd == XM_MACHO_LC_LOAD_WEAK_DYLIB || lc.cmd == XM_MACHO_LC_REEXPORT_DYLIB) {
diff --git a/core/src/xmake/binutils/macho/rpath.c b/core/src/xmake/binutils/macho/rpath.c
new file mode 100644
index 000000000..e67c2edfd
--- /dev/null
+++ b/core/src/xmake/binutils/macho/rpath.c
@@ -0,0 +1,146 @@
+/*!A cross-platform build utility based on Lua
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ * http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ *
+ * Copyright (C) 2015-present, Xmake Open Source Community.
+ *
+ * @author ruki
+ * @file rpath.c
+ *
+ */
+
+/* //////////////////////////////////////////////////////////////////////////////////////
+ * trace
+ */
+#define TB_TRACE_MODULE_NAME "rpath_macho"
+#define TB_TRACE_MODULE_DEBUG (0)
+
+/* //////////////////////////////////////////////////////////////////////////////////////
+ * includes
+ */
+#include "prefix.h"
+
+/* //////////////////////////////////////////////////////////////////////////////////////
+ * implementation
+ */
+tb_bool_t xm_binutils_macho_rpath_list(tb_stream_ref_t istream, tb_hize_t base_offset, lua_State *lua) {
+ tb_assert_and_check_return_val(istream && lua, tb_false);
+
+ // read Mach-O header
+ xm_macho_header_t header;
+ if (!tb_stream_seek(istream, base_offset)) {
+ return tb_false;
+ }
+ if (!tb_stream_bread(istream, (tb_byte_t*)&header, sizeof(header))) {
+ return tb_false;
+ }
+
+ tb_bool_t swap = tb_false;
+ tb_bool_t is64 = tb_false;
+
+ // check magic
+ if (header.magic == XM_MACHO_MAGIC_32) {
+ is64 = tb_false;
+ } else if (header.magic == XM_MACHO_MAGIC_32_BE) {
+ is64 = tb_false;
+ swap = tb_true;
+ } else if (header.magic == XM_MACHO_MAGIC_64) {
+ is64 = tb_true;
+ } else if (header.magic == XM_MACHO_MAGIC_64_BE) {
+ is64 = tb_true;
+ swap = tb_true;
+ } else {
+ return tb_false; // Not a Mach-O file
+ }
+
+ // re-read header for 64-bit if needed, or just use 32-bit part and skip reserved
+ tb_uint32_t ncmds = 0;
+
+ if (is64) {
+ xm_macho_header_64_t header64;
+ if (!tb_stream_seek(istream, base_offset)) {
+ return tb_false;
+ }
+ if (!tb_stream_bread(istream, (tb_byte_t*)&header64, sizeof(header64))) {
+ return tb_false;
+ }
+ xm_binutils_macho_swap_header_64(&header64, swap);
+ ncmds = header64.ncmds;
+ } else {
+ xm_binutils_macho_swap_header_32(&header, swap);
+ ncmds = header.ncmds;
+ }
+
+ // skip header to reach load commands
+ tb_size_t header_size = is64 ? sizeof(xm_macho_header_64_t) : sizeof(xm_macho_header_t);
+ if (!tb_stream_seek(istream, base_offset + header_size)) {
+ return tb_false;
+ }
+
+ tb_size_t result_count = 0;
+
+ // iterate load commands
+ for (tb_uint32_t i = 0; i < ncmds; i++) {
+ xm_macho_load_command_t lc;
+ tb_hize_t current_cmd_offset = tb_stream_offset(istream);
+
+ if (!tb_stream_bread(istream, (tb_byte_t*)&lc, sizeof(lc))) {
+ return tb_false;
+ }
+ xm_binutils_macho_swap_load_command(&lc, swap);
+
+ // check for LC_RPATH
+ if (lc.cmd == XM_MACHO_LC_RPATH) {
+
+ xm_macho_rpath_command_t rc;
+ if (tb_stream_seek(istream, current_cmd_offset)) {
+ if (tb_stream_bread(istream, (tb_byte_t*)&rc, sizeof(rc))) {
+ xm_binutils_macho_swap_rpath_command(&rc, swap);
+
+ tb_uint32_t name_offset = rc.path_offset;
+ if (name_offset < lc.cmdsize) {
+ // name is at current_cmd_offset + name_offset
+ if (tb_stream_seek(istream, current_cmd_offset + name_offset)) {
+ tb_char_t rpath[TB_PATH_MAXN];
+ tb_size_t max_len = lc.cmdsize - name_offset;
+ if (max_len > sizeof(rpath) - 1) max_len = sizeof(rpath) - 1;
+
+ tb_size_t pos = 0;
+ tb_byte_t c;
+ while (pos < max_len) {
+ if (!tb_stream_bread(istream, &c, 1)) break;
+ if (c == 0) break;
+ rpath[pos++] = (tb_char_t)c;
+ }
+ rpath[pos] = '\0';
+
+ if (pos > 0) {
+ lua_pushinteger(lua, result_count + 1);
+ lua_pushstring(lua, rpath);
+ lua_settable(lua, -3);
+ result_count++;
+ }
+ }
+ }
+ }
+ }
+ }
+
+ // move to next command
+ if (!tb_stream_seek(istream, current_cmd_offset + lc.cmdsize)) {
+ break;
+ }
+ }
+
+ return tb_true;
+}
diff --git a/core/src/xmake/binutils/macho/prefix.h b/core/src/xmake/binutils/macho/prefix.h
index 4216bbdd1..612033029 100644
--- a/core/src/xmake/binutils/macho/prefix.h
+++ b/core/src/xmake/binutils/macho/prefix.h
@@ -29,10 +29,10 @@
/* //////////////////////////////////////////////////////////////////////////////////////
* macros
*/
-#define XM_MACHO_MAGIC_32 0xfeedface /* MH_MAGIC - little endian */
-#define XM_MACHO_MAGIC_64 0xfeedfacf /* MH_MAGIC_64 - little endian */
-#define XM_MACHO_MAGIC_32_BE 0xcefaedfe /* MH_CIGAM - big endian */
-#define XM_MACHO_MAGIC_64_BE 0xcffaedfe /* MH_CIGAM_64 - big endian */
+#define XM_MACHO_MAGIC_32 0xfeedface // MH_MAGIC - little endian
+#define XM_MACHO_MAGIC_64 0xfeedfacf // MH_MAGIC_64 - little endian
+#define XM_MACHO_MAGIC_32_BE 0xcefaedfe // MH_CIGAM - big endian
+#define XM_MACHO_MAGIC_64_BE 0xcffaedfe // MH_CIGAM_64 - big endian
#define XM_MACHO_MAGIC_FAT 0xcafebabe
#define XM_MACHO_CPU_TYPE_X86 7
@@ -52,6 +52,7 @@
#define XM_MACHO_LC_SYMTAB 0x2
#define XM_MACHO_LC_LOAD_DYLIB 0xc
#define XM_MACHO_LC_ID_DYLIB 0xd
+#define XM_MACHO_LC_RPATH (0x1c | 0x80000000)
#define XM_MACHO_LC_LOAD_WEAK_DYLIB (0x18 | 0x80000000)
#define XM_MACHO_LC_REEXPORT_DYLIB (0x1f | 0x80000000)
#define XM_MACHO_LC_BUILD_VERSION 0x32
@@ -98,6 +99,13 @@ typedef struct __xm_macho_header_64_t {
tb_uint32_t reserved;
} __tb_packed__ xm_macho_header_64_t;
+typedef struct __xm_macho_rpath_command_t {
+ tb_uint32_t cmd;
+ tb_uint32_t cmdsize;
+ tb_uint32_t path_offset;
+} __tb_packed__ xm_macho_rpath_command_t;
+#include "tbox/prefix/packed.h"
+
typedef struct __xm_macho_segment_command_t {
tb_uint32_t cmd;
tb_uint32_t cmdsize;
@@ -213,7 +221,7 @@ typedef struct __xm_macho_dylib_command_t {
* inline implementation
*/
-/* byte-swap Mach-O header fields if needed */
+// byte-swap Mach-O header fields if needed
static __tb_inline__ tb_void_t xm_binutils_macho_swap_header_32(xm_macho_header_t *header, tb_bool_t swap) {
if (swap) {
header->magic = tb_bits_swap_u32(header->magic);
@@ -226,7 +234,7 @@ static __tb_inline__ tb_void_t xm_binutils_macho_swap_header_32(xm_macho_header_
}
}
-/* byte-swap Mach-O header 64 fields if needed */
+// byte-swap Mach-O header 64 fields if needed
static __tb_inline__ tb_void_t xm_binutils_macho_swap_header_64(xm_macho_header_64_t *header, tb_bool_t swap) {
if (swap) {
header->magic = tb_bits_swap_u32(header->magic);
@@ -240,7 +248,7 @@ static __tb_inline__ tb_void_t xm_binutils_macho_swap_header_64(xm_macho_header_
}
}
-/* byte-swap load command fields if needed */
+// byte-swap load command fields if needed
static __tb_inline__ tb_void_t xm_binutils_macho_swap_load_command(xm_macho_load_command_t *lc, tb_bool_t swap) {
if (swap) {
lc->cmd = tb_bits_swap_u32(lc->cmd);
@@ -248,7 +256,7 @@ static __tb_inline__ tb_void_t xm_binutils_macho_swap_load_command(xm_macho_load
}
}
-/* byte-swap dylib command fields if needed */
+// byte-swap dylib command fields if needed
static __tb_inline__ tb_void_t xm_binutils_macho_swap_dylib_command(xm_macho_dylib_command_t *dc, tb_bool_t swap) {
if (swap) {
dc->cmd = tb_bits_swap_u32(dc->cmd);
@@ -260,7 +268,16 @@ static __tb_inline__ tb_void_t xm_binutils_macho_swap_dylib_command(xm_macho_dyl
}
}
-/* byte-swap symtab command fields if needed */
+// byte-swap rpath command fields if needed
+static __tb_inline__ tb_void_t xm_binutils_macho_swap_rpath_command(xm_macho_rpath_command_t *rc, tb_bool_t swap) {
+ if (swap) {
+ rc->cmd = tb_bits_swap_u32(rc->cmd);
+ rc->cmdsize = tb_bits_swap_u32(rc->cmdsize);
+ rc->path_offset = tb_bits_swap_u32(rc->path_offset);
+ }
+}
+
+// byte-swap symtab command fields if needed
static __tb_inline__ tb_void_t xm_binutils_macho_swap_symtab_command(xm_macho_symtab_command_t *cmd, tb_bool_t swap) {
if (swap) {
cmd->cmd = tb_bits_swap_u32(cmd->cmd);
@@ -272,7 +289,7 @@ static __tb_inline__ tb_void_t xm_binutils_macho_swap_symtab_command(xm_macho_sy
}
}
-/* byte-swap nlist 32 fields if needed */
+// byte-swap nlist 32 fields if needed
static __tb_inline__ tb_void_t xm_binutils_macho_swap_nlist_32(xm_macho_nlist_t *nlist, tb_bool_t swap) {
if (swap) {
nlist->strx = tb_bits_swap_u32(nlist->strx);
@@ -281,7 +298,7 @@ static __tb_inline__ tb_void_t xm_binutils_macho_swap_nlist_32(xm_macho_nlist_t
}
}
-/* byte-swap nlist 64 fields if needed */
+// byte-swap nlist 64 fields if needed
static __tb_inline__ tb_void_t xm_binutils_macho_swap_nlist_64(xm_macho_nlist_64_t *nlist, tb_bool_t swap) {
if (swap) {
nlist->strx = tb_bits_swap_u32(nlist->strx);